;	         A counter-example
; to show that an implementation of dynamic variables in Petite Chez
; scheme (aka parameterize) leads to the unexpected behavior in the
; presence of delimited continuations (shift/reset)
;
; This code is tested with Petite Chez Scheme

; Danvy/Filinski implementation of shift/reset, excerpted from Scheme48
; distribution.
(load "shift-reset.scm")		



;------------------------------------------------------------------------
; For simplicity and consistency with other tests, we use only one
; dynvar, so we don't bother with naming it explicitly. For
; compatibility with the SML code (see *.sml files in this directory)
; we don't make dlet a macro -- hence dlet takes a thunk as the second
; argument.


; dereference the (implicit) dynamic variable
; Just invoke (dref), which will be bound via parameterize
; (define (dref) ...)
(define dref (make-parameter #f))


; Evaluate THUNK in the dynamic environment extended with the binding of the
; (implicit) dynvar to V
(define (dlet v thunk)
  (parameterize ((dref v)) (thunk)))

; The first test
(define (test1)
  (dlet 1
    (lambda ()
      (list (dref) (dref)
      (dlet 2 
	(lambda ()
	  (list (dref) (dref))))))))

(display (test1)) (newline)
; (1 1 (2 2))


; The counter-example itself
; Its code is the same as in the other dynvar counter-examples

; A sample function to print the current dynamic environment, twice
; (just for the heck of it).
(define (tf title)
  (display title) (newline)
  (display (dref)) (newline)
  (display (dref)) (newline))

; The first test: we capture a delimited continuation that 
; DOES include `dlet 2'
(define (test2)
  (newline) (display "test2") (newline)
  (dlet 1
    (lambda ()
      (let ((f 
	      (reset
		(dlet 2 
		  (lambda ()
		    ((shift f (begin (tf "inside shift") f)) "inside again")
		    )))))
	(tf "outside")
	(f tf)
	(tf "outside again")
))))
(test2)

; The second test: we now capture a delimited continuation that does NOT
; include `dlet 2'
(define (test3)
  (newline) (display "test3") (newline)
  (dlet 1
    (lambda ()
      (let ((f 
		(dlet 2 
		  (lambda ()
	      (reset
		    ((shift f (begin (tf "inside shift") f)) "inside again")
		    )))))
	(tf "outside")
	(f tf)
	(tf "outside again")
))))
(test3)

; We observe, test2 and test3 print the same results. 
; But that is quite unexpected and at odds with the combined direct
; semantics of dynamic variables and delimited continuations.
; In test3, the dynamic environment in the case "inside again"
; (that is, when the captured continuation is reinstated)
; must be 1 rather than 2 because `dlet 2' is NOT included within the
; captured continuation.
